import json
from datetime import datetime, timedelta

import jwt
from django.http.response import JsonResponse
from mock import Mock
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.test import APITestCase


class APITestCaseWrapper(APITestCase):
    default_response_headers = {'Content-Type': 'application/json'}

    def __init__(self, *args, **kwargs):
        super(APITestCaseWrapper, self).__init__(*args, **kwargs)

        def to_json(response):
            self.assertEqual(response['Content-Type'], 'application/json')
            response._json = getattr(response, '_json', None) or json.loads(response.content)
            return response._json

        Response.json = to_json
        JsonResponse.json = to_json

    def mocked_response(self, status_code, json_return_value=None, content=None, response_headers=None):
        response = Mock()

        response.status_code = status_code
        response.json.return_value = json_return_value
        response.content = content
        response.headers = response_headers or self.default_response_headers

        return response

    def generate_auth_header(self, **kwargs):
        token = self.jwt_encode(**kwargs)
        self.client.credentials(HTTP_AUTHORIZATION=token)

    @staticmethod
    def jwt_encode(jwt_secret='', algorithm='HS256', ttl=None, **kwargs):
        data = kwargs.copy()
        data['iat'] = datetime.utcnow()
        data['exp'] = datetime.utcnow() + (ttl or timedelta(minutes=10))
        return jwt.encode(data, jwt_secret, algorithm).decode('utf-8')

    def assertExists(self, cls, **kwargs):
        self.assertTrue(cls.objects.filter(**kwargs).exists(), '%s %s does not exist.' % (cls, kwargs))

    def assertDoesntExist(self, cls, **kwargs):
        self.assertFalse(cls.objects.filter(**kwargs).exists(), '%s %s exists, but expected not to.' % (cls, kwargs))

    @staticmethod
    def format_datetime(dt):
        return JSONRenderer().render(dt).strip('"')



"""
Django REST framework handles the following exceptions, and automatically generates error responses:
  - Subclasses of APIException raised inside REST framework,
  - Django's Http404 exception,
  - and Django's PermissionDenied exception.

More info: http://www.django-rest-framework.org/api-guide/exceptions/

We capture them here, and reformat responses to our desired style.
"""
import logging

from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions, status
from rest_framework.response import Response
from rest_framework.views import exception_handler

from django_on_chain import SETTINGS
from django_on_chain.exceptions import ServiceError

logger = logging.getLogger(__name__)


def reformat_error_handler(e, context):
    if isinstance(e, ServiceError):
        content = {
            'type': e.type,
            'message': e.message,
        }
        if e.errors is not None:
            content['errors'] = e.errors
        return Response(content, status=e.suggested_http_status or status.HTTP_500_INTERNAL_SERVER_ERROR)

    response = exception_handler(e, context)  # Call default exception handler to get the standard error response

    if response is None:
        logger.exception('Uncaught exception', extra={
            'request': context['request'],
            'exception': e,
        })

    reformat_errors = getattr(context.get('view', object()), 'reformat_errors', SETTINGS['REFORMAT_ALL_ERRORS'])

    if not reformat_errors:
        return response

    if response is None:
        return Response({
            'type': 'server_error',
            'message': _('Unknown error; please try again later'),
        }, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    if response.status_code == status.HTTP_400_BAD_REQUEST:
        message = _('Invalid request')
        if isinstance(e, exceptions.ValidationError):
            message += ': {}'.format(e.detail)
        response.data = {
            'type': 'bad_request',
            'message': message,
        }
    elif response.status_code == status.HTTP_404_NOT_FOUND:
        response.data = {
            'type': 'not_found',
            'message': _('Resource was not found'),
        }
    elif response.status_code == status.HTTP_401_UNAUTHORIZED:
        response.data = {
            'type': 'unauthorized',
            'message': _('Request is not authorized'),
        }
    elif response.status_code == status.HTTP_403_FORBIDDEN:
        response.data = {
            'type': 'access_forbidden',
            'message': _('You do not have permission to perform this action.'),
        }
    elif isinstance(e, exceptions.APIException):
        response.data = {
            'type': e.default_code,
            'message': force_text(e.default_detail),
        }

    return response



from django.utils.translation import ugettext_lazy as _
from rest_framework import status


class ServiceError(Exception):
    """
    Base class for microservice errors

    Typically a Http response is generated from this.
    """
    def __init__(self, type, message, errors=None, suggested_http_status=None):
        super(ServiceError, self).__init__(message)
        self.type = type
        self.message = message
        self.errors = errors
        self.suggested_http_status = suggested_http_status


class BadRequestError(ServiceError):
    """
    Is raised when an invalid request comes from client
    """
    def __init__(self, type, message, errors=None, suggested_http_status=None):
        super(BadRequestError, self).__init__(type, message, errors,
                                              suggested_http_status or status.HTTP_400_BAD_REQUEST)


class NotFoundError(ServiceError):
    """
    Is raised when a requested entity does not exist
    """
    def __init__(self,
                 type='not_found',
                 message=_('Resource was not found'),
                 errors=None):
        super(NotFoundError, self).__init__(type, message, errors, status.HTTP_404_NOT_FOUND)


class UnauthorizedError(ServiceError):
    """
    Is raised when a request has not valid authentication credentials
    """
    def __init__(self,
                 type='unauthorized',
                 message=_('Request is not authorized'),
                 errors=None):
        super(UnauthorizedError, self).__init__(type, message, errors, status.HTTP_401_UNAUTHORIZED)


class AccessForbiddenError(ServiceError):
    """
    Is raised when the user is logged in, but has not permission to perform the request
    """
    def __init__(self,
                 type='access_forbidden',
                 message=_('You do not have permission to perform this action.'),
                 errors=None):
        super(AccessForbiddenError, self).__init__(type, message, errors, status.HTTP_403_FORBIDDEN)


class ServerError(ServiceError):
    """
    Is raised when an internal server error occurs
    """
    def __init__(self,
                 type='server_error',
                 message=_('Unknown error; please try again later'),
                 errors=None,
                 suggested_http_status=None):
        super(ServerError, self).__init__(type, message, errors,
                                          suggested_http_status or status.HTTP_500_INTERNAL_SERVER_ERROR)
